from bs4 import BeautifulSoup
import os
from . import dem, hdf
from datetime import datetime


class Model:
    def __init__(self, library_file_path):
        self.library = library_file_path
        with open(library_file_path) as f:
            self.tree = BeautifulSoup(f, 'html.parser')
        self.project_file = self.get('ProjectFile')
        self.catchment_name = self.get('CatchmentName')
        self.mask = self.get_path('MaskFileName')
        self.veg_map = self.get_path('VegMap')
        self.lake_map = self.get_path('LakeMap')
        self.precip_map = self.get_path('PrecipMap')
        self.pe_map = self.get_path('PeMap')
        self.veg_details = self.tree.find('VegetationDetails')
        self.soil_properties = self.tree.find('SoilProperties')
        self.soil_details = self.tree.find('SoilDetails')
        self.initial_conditions = self.get('InitialConditions')
        self.precipitation_time_series = self.get_path('PrecipitationTimeSeriesData')
        self.precipitation_time_step = self.tree.find('PrecipitationTimeStep')
        self.evaporation_time_series = self.get_path('EvaporationTimeSeriesData')
        self.evaporation_time_step = self.tree.find('EvaporationTimeStep')
        self.max_temp_time_series = self.get_path('MaxTempTimeSeriesData')
        self.min_temp_time_series = self.get_path('MinTempTimeSeriesData')
        self.start_day = self.get('StartDay')
        self.start_month = self.get('StartMonth')
        self.start_year = self.get('StartYear')
        self.start_date = datetime(self.start_year, self.start_month, self.start_day)
        self.end_day = self.get('EndDay')
        self.end_month = self.get('EndMonth')
        self.end_year = self.get('EndYear')
        self.end_date = datetime(self.end_year, self.end_month, self.end_day) \
            if self.end_day and self.end_month and self.end_year else None
        self.river_grid_squares_accumulated = self.get('RiverGridSquaresAccumulated')
        self.drop_from_grid_to_channel_depth = self.get('DropFromGridToChannelDepth')
        self.minimum_drop_between_channels = self.get('MinimumDropBetweenChannels')
        self.regular_time_step = self.get('RegularTimestep')
        self.increasing_time_step = self.get('IncreasingTimestep')
        self.simulated_discharge_time_step = self.get('SimulatedDischargeTimestep')
        self.snow_melt_degree_day_factor = self.get('SnowmeltDegreeDayFactor')
        self.snow_melt_degree_day_factor = self.get('SnowmeltDegreeDayFactor')
        self.srid = self.get('SRID')

        self.dem = dem.Dem(self.get_path('DEMMeanFileName'))
        self.hdf = hdf.Hdf(self.path('output_{}_shegraph.h5'.format(self.catchment_name)),
                           model=self)

    def get(self, name):
        value = self.tree.find(name.lower())
        if value is None:
            return
        else:
            value = value.get_text(strip=True)
        if value is None:
            return
        elif value.isdigit():
            return int(value)
        else:
            return value

    def path(self, name):
        return os.path.join(os.path.dirname(self.library), str(name)) if name is not None else name

    def get_path(self, name):
        return self.path(self.get(name))







# preparing the library file to go into SHETRAN e.g. H:\PhD\PythonLessons\gettingItIntoSHETRAN\s1.xml
# please ensure that there are no backslashes in the file paths as theres are escape characters in python. Use forward slashes instead!
import os


def make_lib_file(run):

    print("Making library file")
    lib_file_path = os.path.join(run.directory, "LibraryFile.xml")
    project_file = "{}ProjectFile".format(run.gauge_id)
    dem_file = os.path.basename(run.outputs.dem)
    min_dem_file = os.path.basename(run.outputs.min_dem)
    mask_files = os.path.basename(run.outputs.mask)
    pe_file = os.path.basename(run.outputs.pe)
    ppt_file = os.path.basename(run.outputs.rain)
    veg_file = os.path.basename(run.outputs.veg)
    soil_file = os.path.basename(run.outputs.soil)
    # if run.resolution == '500m':
    lake_file = os.path.basename(run.outputs.lakes)
    # else:
    #     lake_file = False
    ppt_time_series = os.path.basename(run.outputs.rain_timeseries)
    pe_time_series = os.path.basename(run.outputs.pe_timeseries)
    max_temp_time_series = os.path.basename(run.outputs.max_temp_timeseries)
    min_temp_time_series = os.path.basename(run.outputs.min_temp_timeseries)
    simulated_discharge_file = os.path.basename(run.outputs.simulated_discharge)

    library_file = open(lib_file_path, "w")

    # input files
    with open("code_key_uk_soil_four_layers.txt", "r") as f:
        code_key_uk_soil_four_layers = f.readlines()

    # Variables
    overland_flow_roughness_coefficient = 1  # 0.1-10
    initial_conditions = 0
    precipitation_time_step = 24
    pe_time_step = 24
    if run.start_date[8] == 0:
        start_day = run.start_date[9:]
    else:
        start_day = run.start_date[8:]

    if run.start_date[5] == 0:
        start_month = run.start_date[6:-3]
    else:
        start_month = run.start_date[5:-3]
    start_year = run.start_date[:-6]

    if run.end_date[8] == 0:
        end_day = run.end_date[9:]
    else:
        end_day = run.end_date[8:]

    if run.end_date[5] == 0:
        end_month = run.end_date[6:-3]
    else:
        end_month = run.end_date[5:-3]
    end_year = run.end_date[:-6]

    river_grid_squares_accumulated = 2  # min 2 - 200
    drop_from_grid_to_channel_depth = 2  # 2-20
    minimum_drop_between_channels = 0.5  # 0.1- 10 only 10 in steep catchment
    rock_depth = 5  # (1-20)
    snowmelt_degree_day_factor = 0.0002
    # ae/pe 0.1-1

    # soil parameters
    hg1 = ['HighlyProductiveAquifer, ', '0.3, 0.2, 0.1, 0.01, 5']
    hg2 = ['ModeratelyProductiveAquifer, ', '0.3, 0.2, 0.01, 0.01, 5']
    hg3 = ['LowProductivityAquifer, ', '0.3, 0.2, 0.001, 0.01, 5']
    hg4 = ['NoGroundwater, ', '0.3, 0.2, 0.0001, 0.01, 5']

    rock_depth_for_h_p_aquifer = 20

    dr1 = 1
    dr2 = 0.3  # no value"
    dr3 = 0.6
    dr4 = 0.4
    dr5 = 1.2

    text_dep_chg0 = 0.3  # "No information"
    text_dep_chg1 = 0.3
    text_dep_chg2 = 0.5
    text_dep_chg3 = 0.7
    text_dep_chg4 = 1
    text_dep_chg5 = 1.3  # check with steve. I think this should just represent the botom of the column
    text_dep_chg6 = 0.4
    text_dep_chg7 = 0.9

    # EUcode, soil layer, soil type, saturated water content, Residual Water Content, Saturated Conductivity (m/day), vanGenuchten- alpha (cm-1), vanGenuchten-n
    text_srf_dom0 = ['1, Peat, ', '0.766, 0.010, 8, 0.0130, 1.2039']  # "1, NoInformation"
    text_srf_dom9 = ['1, Peat, ', '0.766, 0.010, 8, 0.0130, 1.2039']
    text_srf_dom1 = ['1, Coarse(18%:clayAnd:65%sand), ', '0.403, 0.025, 60.000, 0.0383, 1.3774']
    text_srf_dom2 = ['1, Medium(18%:clay:35%And:15%sandOr18%:clayAnd15%:sand:65%), ',
                   '0.439, 0.010, 12.061, 0.0314, 1.1804']
    text_srf_dom3 = ['1, MediumFine(:35%clayand:15%sand), ', '0.430, 0.010, 2.272, 0.0083, 1.2539']
    text_srf_dom4 = ['1, Fine(35%:clay:60%), ', '0.520, 0.010, 24.800, 0.0367, 1.1012']
    text_srf_dom5 = ['1, VeryFine(clay:60%), ', '0.614, 0.010, 15.000, 0.0265, 1.1033']

    # EUcode, soil layer, soil type, saturated water content, Residual Water Content, Saturated Conductivity (m/day), vanGenuchten- alpha (cm-1), vanGenuchten-n
    text_sub_dom0 = ['2, Peat, ', '0.766, 0.010, 8, 0.0130, 1.2039']  # "2, NoInformation"
    text_sub_dom9 = ['2, Peat, ', '0.766, 0.010, 8, 0.0130, 1.2039']
    text_sub_dom1 = ['2, Coarse(18%:clayAnd:65%sand), ', '0.366, 0.025, 70.000, 0.0430, 1.5206']
    text_sub_dom2 = ['2, Medium(18%:clay:35%And::15%sandOr18%:clayAnd15%:sand:65%), ',
                   '0.329, 0.010, 10.755, 0.0249, 1.1689']
    text_sub_dom3 = ['2, MediumFine(:35%clayand:15%sand), ', '0.412, 0.010, 4.000, 0.0082, 1.2179']
    text_sub_dom4 = ['2, Fine(35%:clay:60%), ', '0.481, 0.010, 8.500, 0.0198, 1.0861']
    text_sub_dom5 = ['2, VeryFine(clay:60%), ', '0.538, 0.010, 8.235, 0.0168, 1.0730']

    # write file
    library_file.write("<?xml version=""1.0""?><ShetranInput>\n")
    library_file.write("<ProjectFile>" + project_file + "</ProjectFile>\n")
    library_file.write("<CatchmentName>{}</CatchmentName>\n".format(run.gauge_id))
    library_file.write("<DEMMeanFileName>" + dem_file + "</DEMMeanFileName>\n")
    library_file.write("<DEMminFileName>" + min_dem_file + "</DEMMinFileName>\n")
    library_file.write("<MaskFileName>" + mask_files + "</MaskFileName>\n")
    library_file.write("<VegMap>" + veg_file + "</VegMap>\n")
    library_file.write("<SoilMap>" + soil_file + "</SoilMap>\n")
    # if lake_file:
    library_file.write("<LakeMap>" + lake_file + "</LakeMap>\n")
    library_file.write("<PrecipMap>" + ppt_file + "</PrecipMap>\n")
    library_file.write("<PeMap>" + pe_file + "</PeMap>\n")
    library_file.write(
        "<VegetationDetails><VegetationDetail>Veg Type #, Vegetation Type, Canopy storage capacity (mm), Leaf area index, Maximum rooting depth(m), AE/PE at field capacity</VegetationDetail>\n")

    # veg key:
    # <VegetationDetail>1, Grass, 1.5, 6, 1, 0.6</VegetationDetail>\n

    library_file.write("<VegetationDetail>1, Arable, 1.0, 0.8, 0.8, 0.6</VegetationDetail>\n")
    library_file.write("<VegetationDetail>2, BareGround, 0, 0, 0.1, 0.4</VegetationDetail>\n")
    library_file.write("<VegetationDetail>3, Grass, 1.5, 1, 1.0, 0.6</VegetationDetail>\n")
    library_file.write("<VegetationDetail>4, DeciduousForest, 5, 1, 1.6, 1.0</VegetationDetail>\n")
    library_file.write("<VegetationDetail>5, EvergreenForest, 5, 1, 2.0, 1.0</VegetationDetail>\n")
    library_file.write("<VegetationDetail>6, Shrub, 1.5, 1, 1.0, 0.4</VegetationDetail>\n")
    library_file.write("<VegetationDetail>7, Urban, 0.3, 0.3, 0.5, 0.4</VegetationDetail>\n")
    library_file.write("</VegetationDetails>\n")

    library_file.write(
        "<SoilDetails><SoilDetail>Soil Category, Soil Layer, Soil Type, Depth at base of layer (m), Saturated Water Content, Residual Water Content, Saturated Conductivity (m/day), vanGenuchten- alpha (cm-1), vanGenuchten-n</SoilDetail>\n")
    # loop over soil column types- format:
    # <SoilDetail>1, 1, SiltLoam(10%Sand:10%Clay), 1, 0.452, 0.093, 0.163, 5.15E-03, 1.681</SoilDetail>\n

    for line in code_key_uk_soil_four_layers:
        lineList = line.rstrip().split(" ")
        # print lineList
        # get she_code
        she_code = lineList[1].rstrip()
        # data or no data?
        if she_code == "-9999":
            print("")
        else:
            # getting the first element of the line i.e. the long colon code
            code_list = lineList[0].rstrip().split(":")
            # print code_list
            hg = code_list[0]
            dr = code_list[1]
            text_dep_chg = code_list[2]
            text_srf_dom = code_list[3]
            text_sub_dom = code_list[4]

            if float(eval("dr" + str(dr))) <= float(eval("text_dep_chg" + str(text_dep_chg))):  # 1 layer plus rock layer
                if float(dr) == 1:
                    second_bit = dr1
                elif float(dr) == 2:
                    second_bit = dr2
                elif float(dr) == 3:
                    second_bit = dr3
                elif float(dr) == 4:
                    second_bit = dr4
                elif float(dr) == 5:
                    second_bit = dr5
                else:
                    print("piddle on aisle dr")

                if float(text_srf_dom) == 0:
                    first_bit = text_srf_dom0[0]
                    third_bit = text_srf_dom0[1]
                elif float(text_srf_dom) == 9:
                    first_bit = text_srf_dom9[0]
                    third_bit = text_srf_dom9[1]
                elif float(text_srf_dom) == 1:
                    first_bit = text_srf_dom1[0]
                    third_bit = text_srf_dom1[1]
                elif float(text_srf_dom) == 2:
                    first_bit = text_srf_dom2[0]
                    third_bit = text_srf_dom2[1]
                elif float(text_srf_dom) == 3:
                    first_bit = text_srf_dom3[0]
                    third_bit = text_srf_dom3[1]
                elif float(text_srf_dom) == 4:
                    first_bit = text_srf_dom4[0]
                    third_bit = text_srf_dom4[1]
                elif float(text_srf_dom) == 5:
                    first_bit = text_srf_dom5[0]
                    third_bit = text_srf_dom5[1]
                else:
                    print("piddle on aisle text_srf_dom")
                # Added in that if it is a highly productive aquifer, the depth of it is 20m not 5m
                if float(hg) == 1:
                    first_bit2 = hg1[0]
                    third_bit2 = hg1[1]
                    rock_depth = 20
                elif float(hg) == 2:
                    first_bit2 = hg2[0]
                    third_bit2 = hg2[1]
                elif float(hg) == 3:
                    first_bit2 = hg3[0]
                    third_bit2 = hg3[1]
                elif float(hg) == 4:
                    first_bit2 = hg4[0]
                    third_bit2 = hg4[1]
                else:
                    print('piddle on aisle hg')
                # 1 layer topsoil
                library_file.write(
                    "<SoilDetail>" + she_code + ", " + first_bit + str(second_bit) + ", " + third_bit + "</SoilDetail>\n")
                # 2 layer rock
                library_file.write("<SoilDetail>" + she_code + ", 2, " + first_bit2 + str(
                    second_bit + rock_depth) + ", " + third_bit2 + "</SoilDetail>\n")
            else:
                # 3 layers	including rock
                # 1st layer
                if float(text_srf_dom) == 0:
                    first_bit = text_srf_dom0[0]
                    third_bit = text_srf_dom0[1]
                elif float(text_srf_dom) == 9:
                    first_bit = text_srf_dom9[0]
                    third_bit = text_srf_dom9[1]
                elif float(text_srf_dom) == 1:
                    first_bit = text_srf_dom1[0]
                    third_bit = text_srf_dom1[1]
                elif float(text_srf_dom) == 2:
                    first_bit = text_srf_dom2[0]
                    third_bit = text_srf_dom2[1]
                elif float(text_srf_dom) == 3:
                    first_bit = text_srf_dom3[0]
                    third_bit = text_srf_dom3[1]
                elif float(text_srf_dom) == 4:
                    first_bit = text_srf_dom4[0]
                    third_bit = text_srf_dom4[1]
                elif float(text_srf_dom) == 5:
                    first_bit = text_srf_dom5[0]
                    third_bit = text_srf_dom5[1]
                else:
                    print("piddle on aisle text_srf_dom")

                if float(text_dep_chg) == 0:
                    second_bit = text_dep_chg0
                elif float(text_dep_chg) == 1:
                    second_bit = text_dep_chg1
                elif float(text_dep_chg) == 2:
                    second_bit = text_dep_chg2
                elif float(text_dep_chg) == 3:
                    second_bit = text_dep_chg3
                elif float(text_dep_chg) == 4:
                    second_bit = text_dep_chg4
                elif float(text_dep_chg) == 5:
                    second_bit = text_dep_chg5
                elif float(text_dep_chg) == 6:
                    second_bit = text_dep_chg6
                elif float(text_dep_chg) == 7:
                    second_bit = text_dep_chg7
                else:
                    print("piddle on aisle text_dep_chg")

                # 2nd layer

                if float(text_sub_dom) == 0:
                    first_bit2 = text_sub_dom0[0]
                    third_bit2 = text_sub_dom0[1]
                elif float(text_sub_dom) == 9:
                    first_bit2 = text_sub_dom9[0]
                    third_bit2 = text_sub_dom9[1]
                elif float(text_sub_dom) == 1:
                    first_bit2 = text_sub_dom1[0]
                    third_bit2 = text_sub_dom1[1]
                elif float(text_sub_dom) == 2:
                    first_bit2 = text_sub_dom2[0]
                    third_bit2 = text_sub_dom2[1]
                elif float(text_sub_dom) == 3:
                    first_bit2 = text_sub_dom3[0]
                    third_bit2 = text_sub_dom3[1]
                elif float(text_sub_dom) == 4:
                    first_bit2 = text_sub_dom4[0]
                    third_bit2 = text_sub_dom4[1]
                elif float(text_sub_dom) == 5:
                    first_bit2 = text_sub_dom5[0]
                    third_bit2 = text_sub_dom5[1]
                else:
                    print("piddle on aisle text_srf_dom")

                if float(dr) == 1:
                    second_bit2 = dr1
                elif float(dr) == 2:
                    second_bit2 = dr2
                elif float(dr) == 3:
                    second_bit2 = dr3
                elif float(dr) == 4:
                    second_bit2 = dr4
                elif float(dr) == 5:
                    second_bit2 = dr5
                else:
                    print("piddle on aisle dr")

                # rock layer. Added in that if it is a highly productive aquifer, the depth of it is 20m not 5m
                if float(hg) == 1:
                    first_bit3 = hg1[0]
                    third_bit3 = hg1[1]
                    rock_depth = rock_depth_for_h_p_aquifer
                elif float(hg) == 2:
                    first_bit3 = hg2[0]
                    third_bit3 = hg2[1]
                elif float(hg) == 3:
                    first_bit3 = hg3[0]
                    third_bit3 = hg3[1]
                elif float(hg) == 4:
                    first_bit3 = hg4[0]
                    third_bit3 = hg4[1]

                library_file.write(
                    "<SoilDetail>" + she_code + ", " + first_bit + str(second_bit) + ", " + third_bit + "</SoilDetail>\n")
                library_file.write("<SoilDetail>" + she_code + ", " + first_bit2 + str(
                    second_bit2) + ", " + third_bit2 + "</SoilDetail>\n")
                library_file.write("<SoilDetail>" + she_code + ", 3, " + first_bit3 + str(
                    second_bit2 + rock_depth) + ", " + third_bit3 + "</SoilDetail>\n")

    library_file.write("</SoilDetails><OverlandFlowRoughnessCoefficient>" + str(
        overland_flow_roughness_coefficient) + "</OverlandFlowRoughnessCoefficient>\n")
    library_file.write("<InitialConditions>" + str(initial_conditions) + "</InitialConditions>\n")
    library_file.write("<PrecipitationTimeSeriesData>" + ppt_time_series + "</PrecipitationTimeSeriesData>\n")
    library_file.write("<PrecipitationTimeStep>" + str(precipitation_time_step) + "</PrecipitationTimeStep>\n")
    library_file.write("<EvaporationTimeSeriesData>" + pe_time_series + "</EvaporationTimeSeriesData>\n")
    library_file.write("<EvaporationTimeStep>" + str(pe_time_step) + "</EvaporationTimeStep>\n")
    library_file.write("<MaxTempTimeSeriesData>" + max_temp_time_series + "</MaxTempTimeSeriesData>\n")
    library_file.write("<MinTempTimeSeriesData>" + min_temp_time_series + "</MinTempTimeSeriesData>\n")
    library_file.write("<StartDay>" + str(start_day) + "</StartDay>\n")
    library_file.write("<StartMonth>" + str(start_month) + "</StartMonth>\n")
    library_file.write("<StartYear>" + str(start_year) + "</StartYear>\n")
    library_file.write("<EndDay>" + str(end_day) + "</EndDay>\n")
    library_file.write("<EndMonth>" + str(end_month) + "</EndMonth>\n")
    library_file.write("<EndYear>" + str(end_year) + "</EndYear>\n")
    library_file.write(
        "<RiverGridSquaresAccumulated>" + str(river_grid_squares_accumulated) + "</RiverGridSquaresAccumulated>\n")
    library_file.write(
        "<DropFromGridToChannelDepth>" + str(drop_from_grid_to_channel_depth) + "</DropFromGridToChannelDepth>\n")
    library_file.write(
        "<MinimumDropBetweenChannels>" + str(minimum_drop_between_channels) + "</MinimumDropBetweenChannels>\n")
    library_file.write("<SnowmeltDegreeDayFactor>" + str(snowmelt_degree_day_factor) + "</SnowmeltDegreeDayFactor>\n")
    library_file.write("<SimulatedDischargeFile>" + simulated_discharge_file + "</SimulatedDischargeFile>\n")
    library_file.write("<PreviousDischargeFile />\n")
    library_file.write("<MeasuredDischargeFile />\n")
    library_file.write("</ShetranInput>\n")

    print("library file created")
    library_file.close()



import os
import shutil


class Outputs:
    def __init__(self, run):
        self.dem = os.path.join(run.path, "DEM.txt")
        self.min_dem = os.path.join(run.path, "MinDEM.txt")
        self.pe = os.path.join(run.path, "PE.txt")
        self.rain = os.path.join(run.path, "Rain.txt")
        self.veg = os.path.join(run.path, "LandCover.txt")
        self.soil = os.path.join(run.path, "Soil.txt")
        self.lakes = os.path.join(run.path, "Lakes.txt")
        self.rain_timeseries = os.path.join(run.path, "RainTimeSeries.csv")
        self.observed_flow = os.path.join(os.path.join(run.path, "FlowData{}-{}.csv"
                                                       .format(run.start_date.replace("-", "."),
                                                               run.end_date.replace("-", "."))))
        self.mask = os.path.join(run.path, "Mask.txt")
        self.min_temp_timeseries = os.path.join(run.path, "MinTempTimeSeries.csv")
        self.max_temp_timeseries = os.path.join(run.path, "MaxTempTimeSeries.csv")
        self.pe_timeseries = os.path.join(run.path, "PeTimeSeries.csv")
        self.simulated_discharge = "SimulatedDischarge.txt"


class Inputs:
    def __init__(self, run):
        self.dem = "../Inputs/{0}BngMaps/{0}DtmBng.txt".format(run.resolution)
        self.min_dem = "../Inputs/{0}BngMaps/{0}MinDtmBng.txt".format(run.resolution)
        self.soil = "../Inputs/{0}BngMaps/{0}SoilBng.txt".format(run.resolution)
        self.veg = "../Inputs/{0}BngMaps/{0}LcmBng.txt".format(run.resolution)
        self.pe = "../Inputs/{0}BngMaps/{0}PeRainBng.txt".format(run.resolution)
        self.rain = "../Inputs/{0}BngMaps/{0}PeRainBng.txt".format(run.resolution)
        self.lakes = "../Inputs/1kmBngMaps/lakes1kmBNG.txt"
        if run.boundary is None:
            self.mask = "../Inputs/{0}BngMasks/{1}.txt".format(run.resolution, run.gauge_id)
        else:
            self.mask = None


class Run:
    def __init__(
            self,
            resolution,
            start_date,
            end_date,
            gauge_id=None,
            boundary=None,
            path=None,
            overwrite=True
    ):
        """start and end dates must be formatted YYYY-MM-DD and resolution is either '1km', '500m' or '100m'"""
        self.gauge_id = gauge_id
        self.resolution = resolution
        self.resolution.f = 3
        self.start_date = start_date
        self.end_date = end_date
        self.path = os.path.join('outputs', str(self.gauge_id)) if path is None else path


        if os.path.exists(self.path) and overwrite:
            shutil.rmtree(self.path)
            
        if not os.path.exists(path):
            os.mkdir(path)
            
        self.boundary = boundary
        self.outputs = Outputs(self)
        self.inputs = Inputs(self)
        self.lakes_exist = os.path.exists(self.inputs.lakes)

        resolution_lookup = {'1km': 1000,
                             '500m': 500,
                             '100m': 100}

        self.resolution_in_metres = resolution_lookup[self.resolution]





import numpy as np


class Dem:
    def __init__(self, path):
        self.path = path
        with open(self.path) as f:
            lines = f.readlines()
        self.number_of_columns = int(lines[0].split()[1])
        self.number_of_rows = int(lines[1].split()[1])
        self.x_lower_left = float(lines[2].split()[1])
        self.y_lower_left = float(lines[3].split()[1])
        self.cell_size = int(float(lines[4].split()[1]))
        self.x_coordinates = np.array([self.x_lower_left - self.cell_size + self.cell_size/2 + i * self.cell_size
                                       for i in range(self.number_of_columns + 2)])
        self.y_coordinates = np.flip(np.array([self.y_lower_left - self.cell_size + self.cell_size/2 + i * self.cell_size
                                               for i in range(self.number_of_rows + 2)]), axis=0)

    def get_index(self, x, y):

        x_location = list(self.x_coordinates[np.abs(self.x_coordinates - x) == np.min(np.abs(self.x_coordinates - x))])[0]
        y_location = list(self.y_coordinates[np.abs(self.y_coordinates - y) == np.min(np.abs(self.y_coordinates - y))])[0]

        return int(np.where(self.x_coordinates == x_location)[0]), int(np.where(self.y_coordinates == y_location)[0])



import os

class Plan:
    def __init__(self, path):
        self.path = path
        assert os.path.exists(path), 'No such file'
        self.items = []
        with open(self.path) as f:
            for line in f:
                if line == 'item':
                    print(line)
            # self.lines = [line.strip() for line in f.readlines()]


from osgeo import gdal, ogr
import os
import zipfile
import math
import numpy as np

def create(outline: str, resolution: int, output_path: str) -> None:
    """
    Converts a vector catchment outline to a gridded mask at specified resolution
    Mask output is aligned to coordinates 0 0
    Any grid cells intersected by the geometry are designated as within the domain

    :param outline: path to a zipped shapefile, must contain one feature
    :param resolution: resolution of the output mask in metres
    :param output_path: location to save the created tiff file
    """
    zipped_file = zipfile.ZipFile(outline, 'r')
    shp = [f.filename for f in zipped_file.infolist() if f.filename.endswith('shp')][0]
    for f in zipped_file.infolist():
        gdal.FileFromMemBuffer(os.path.join('/vsimem', f.filename), bytes(zipped_file.read(f.filename)))

    upload = ogr.Open(os.path.join('/vsimem', shp))

    layer = upload.GetLayer()

    feature = layer.GetNextFeature()
    geom = feature.GetGeometryRef()
    minX, maxX, minY, maxY = geom.GetEnvelope()

    minX = math.floor(minX / resolution) * resolution
    maxX = math.ceil(maxX / resolution) * resolution
    minY = math.floor(minY / resolution) * resolution
    maxY = math.ceil(maxY / resolution) * resolution

    width = int((maxX - minX) / resolution)
    height = int((maxY - minY) / resolution)

    driver = gdal.GetDriverByName('GTiff')
    target_ds = driver.Create(output_path+'.tif', width, height, 1, gdal.GDT_Float32)
    target_ds.SetGeoTransform((minX, resolution, 0, maxY, 0, -1*resolution))

    no_data = -9999
    band = target_ds.GetRasterBand(1)
    band.WriteArray(np.full([height, width], no_data))
    band.FlushCache()
    band.SetNoDataValue(no_data)
    del target_ds

    target_ds = gdal.Open(output_path + '.tif', gdal.GA_Update)
    gdal.RasterizeLayer(target_ds, [1], layer, burn_values=[1], options=['ALL_TOUCHED=TRUE'])
    target_ds.FlushCache()

    # Convert to ASCII grid
    driver = gdal.GetDriverByName("AAIGrid")
    driver.CreateCopy(output_path, target_ds, 0)

def extract(mask_path: str, input_path: str, output_path: str, resolution: int) -> None:
    """
    Input data and mask must aligned to the same coordinate grid

    :param mask_path: path to mask file in GDAL recognised format
    :param input_path: path to input data in GDAL recognised format
    :param output_path: path to output data in GDAL recognised format
    :param resolution: resolution of
    """
    mask_file = open(mask_path, "r")

    n_cols_line = mask_file.readline()
    n_cols_list = n_cols_line.rstrip().split()
    n_cols = float(n_cols_list[1])

    n_rows_line = mask_file.readline()
    n_rows_list = n_rows_line.rstrip().split()
    n_rows = float(n_rows_list[1])

    bx_line = mask_file.readline()
    bx_list = bx_line.rstrip().split()
    x_lower_left_corner = float(bx_list[1])

    by_line = mask_file.readline()
    by_list = by_line.rstrip().split()
    y_lower_left_corner = float(by_list[1])

    mask_file.close()
    ds = gdal.Open(input_path)

    start_line_number = ds.RasterYSize - y_lower_left_corner / resolution - n_rows
    start_index_number = x_lower_left_corner / resolution

    gdal.Translate(output_path,
                   ds,
                   srcWin=
                   [
                       start_index_number,
                       start_line_number,
                       n_cols,
                       n_rows
                   ],
                   format='AAIGrid')



from setuptools import setup

setup(
    name='shetranio',
    version='0.1.0',
    packages=['shetranio', 'shetranio.setup'],
    url='https://github.com/nclwater/shetranio',
    license='MIT',
    author='Fergus McClean',
    author_email='fergus.mcclean@ncl.ac.uk',
    description='A python package to complement the SHETran hydrological model.',
    install_requires=[
        'numpy',
        'h5py',
        'matplotlib',
        'pandas',
        'beautifulsoup4'
    ]
)



import h5py
from .dem import Dem
import numpy as np
import pandas as pd


class Constant:
    def __init__(self, array):
        self.square = array[:, :, 0]
        self.north_bank = array[:, :, 1]
        self.east_bank = array[:, :, 2]
        self.south_bank = array[:, :, 3]
        self.west_bank = array[:, :, 4]
        self.north_link = array[:, :, 5]
        self.east_link = array[:, :, 6]
        self.south_link = array[:, :, 7]
        self.west_link = array[:, :, 8]

    def get_link(self,direction):
        if direction == 'n':
            return self.north_link
        elif direction == 'e':
            return self.east_link
        elif direction == 's':
            return self.south_link
        elif direction == 'w':
            return self.west_link
        else:
            raise Exception('Please specify a direction from [n,e,s,w]')


class Variable:
    def __new__(cls, hdf, variable_name):
        if variable_name in hdf.variable_names.keys():
            return super(Variable, cls).__new__(cls)
        else:
            return None

    def __init__(self, hdf, variable_name):
        self.name = variable_name
        self.hdf = hdf
        self.variable = hdf.file_variables[hdf.variable_names[variable_name]]
        self.values = self.variable['value']
        self.times = self.variable['time']
        self.time_units = self.times.attrs['units'][0].decode("utf-8")
        if self.hdf.model:
            self.times = pd.date_range(
                start=self.hdf.model.start_date,
                freq='{}H'.format(np.floor(self.times[1]) - np.floor(self.times[0])),
                periods=len(self.times))
            self.time_units = ''
        self.units = self.values.attrs['units'][0].decode("utf-8")
        self.long_name = '{} ({})'.format(variable_names[self.name], self.units)
        self.is_river = False
        self.is_spatial = True
        try:
            self.hdf.variables.append(self)
        except TypeError:
            pass


class RiverVariable(Variable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)
        self.is_river = True


class OverlandFlow(RiverVariable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)

    def get_element(self, element_number):
        return np.abs(self.values[self.hdf.get_element_index(element_number), :, :]).max(axis=0)

    def get_time(self, time_index):
        return np.abs(self.values[:, :, time_index]).max(axis=1)


class SurfaceDepth(RiverVariable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)

    def get_element(self, element_number):
        return self.values[self.hdf.get_element_index(element_number), :]

    def get_time(self, time_index):
        return np.abs(self.values[:, time_index])


class LandVariable(Variable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)

    def get_element(self, element_number):
        index = np.where(self.hdf.number.square == element_number)
        return self.values[index[0][0], index[1][0]]

    def get_time(self, time_index):
        numbers = self.hdf.number.square.flatten()
        values = self.values[:, :, time_index].flatten()
        a = values[numbers != -1][np.argsort(numbers[numbers != -1])]
        a[a == -1] = np.nan
        return a


class SoilMoisture(LandVariable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)

    def get_element(self, element_number, level=0):
        index = np.where(self.hdf.number.square == element_number)
        return self.values[index[0][0], index[1][0], level]

    def get_time(self, time_index, level=0):
        numbers = self.hdf.number.square.flatten()
        values = self.values[:, :, level, time_index].flatten()
        a = values[numbers != -1][np.argsort(numbers[numbers != -1])]
        a[a == -1] = np.nan
        return a


class RainVariable(Variable):
    def __init__(self, hdf, variable_name):
        super().__init__(hdf, variable_name)
        self.is_spatial = False


class Hdf:
    def __init__(self, path, model=None):
        self.path = path
        self.model = model
        self.file = h5py.File(path, 'r', driver='core')
        self.catchment_maps = self.file['CATCHMENT_MAPS']
        self.sv4_elevation = self.catchment_maps['SV4_elevation'][:]
        self.palette1 = self.catchment_maps['palette1']
        self.catchment_spreadsheets = self.file['CATCHMENT_SPREADSHEETS']
        self.sv4_numbering = self.catchment_spreadsheets['SV4_numbering'][:]
        self.element_numbers = np.unique(self.sv4_numbering)[1:]
        self.constants = self.file['CONSTANTS']
        self.centroid = Constant(self.constants['centroid'])
        self.grid_dxy = self.constants['grid_dxy']
        self.number = Constant(self.constants['number'])
        self.land_elements = np.unique(self.number.square)[1:]
        self.river_elements = self.element_numbers[:min(self.land_elements) - 1]
        self.r_span = Constant(self.constants['r_span'])
        self.soil_type = Constant(self.constants['soil_typ'])
        self.spatial1 = Constant(self.constants['spatial1'])
        self.surface_elevation = Constant(self.constants['surf_elv'])
        self.vertical_thickness = Constant(self.constants['vert_thk'])

        self.file_variables = self.file['VARIABLES']
        self.variable_names = dict([(k.split(' ')[-1], k) for k in self.file_variables.keys()])
        self.variables = []
        self.net_rain = RainVariable(self, 'net_rain')
        self.potential_evapotranspiration = LandVariable(self, 'pot_evap')
        self.transpiration = LandVariable(self, 'trnsp')
        self.surface_evaporation = LandVariable(self, 'srf_evap')
        self.evaporation_from_interception = LandVariable(self, 'int_evap')
        self.drainage_from_interception = LandVariable(self, 'drainage')
        self.canopy_storage = LandVariable(self, 'can_stor')
        self.vertical_flows = LandVariable(self, 'v_flow')
        self.snow_depth = LandVariable(self, 'snow_dep')
        self.ph_depth = LandVariable(self, 'ph_depth')
        self.overland_flow = OverlandFlow(self, 'ovr_flow')
        self.surface_depth = SurfaceDepth(self, 'srf_dep')
        self.surface_water_potential = LandVariable(self, 'psi')
        self.soil_moisture = SoilMoisture(self, 'theta')
        self.total_sediment_depth = LandVariable(self, 's_t_dp')
        self.surface_erosion_rate = LandVariable(self, 's_v_er')
        self.sediment_discharge_rate = LandVariable(self, 's_dis')
        self.mass_balance_error = LandVariable(self, 'bal_err')
        self.snow_depth = LandVariable(self, 'snow_dep')
        self.variables = tuple(self.variables)
        self.spatial_variables = tuple([var for var in self.variables if var.is_spatial])
        self.elevations = self.get_elevations()


    def get_element_number(self, dem_file, location):
        d = Dem(dem_file)
        x_index, y_index = d.get_index(location[0], location[1])
        return self.number.square[y_index,x_index]

    def get_element_index(self, element_number):
        return self.element_numbers.tolist().index(element_number)

    def get_elevations(self):
        n = self.number
        e = self.surface_elevation

        numbers = list(n.square)
        elevations = list(e.square)

        for link in ['n', 's', 'w', 'e']:
            numbers.extend(n.get_link(link))
            elevations.extend(e.get_link(link))
        numbers = np.array(numbers).flatten()
        elevations = np.array(elevations).flatten()

        numbers, indices = np.unique(numbers, return_index=True)
        elevations = elevations[indices]

        return elevations[elevations != -1]

    def get_channel_link_number(self, dem_file, location, direction):
        """Returns north-south and east-west channel link numbers"""
        assert direction in ['n', 'e', 's', 'w'], 'Please specify a direction from [n, e, s, w]'

        x_index, y_index = d.get_index(location[0], location[1])

        if direction=='n':
            return self.number.north_link[y_index, x_index]
        elif direction=='e':
            return self.number.east_link[y_index,x_index]
        elif direction=='s':
            return self.number.south_link[y_index,x_index]
        elif direction=='w':
            return self.number.west_link[y_index, x_index]


    def get_channel_link_location(self, dem_file, element_number):
        d = Dem(dem_file)

        if element_number in self.number.north_link:
            index = np.where(self.number.north_link == element_number)
        elif element_number in self.number.east_link:
            index = np.where(self.number.east_link == element_number)
        elif element_number in self.number.south_link:
            index = np.where(self.number.south_link == element_number)
        elif element_number in self.number.west_link:
            index = np.where(self.number.west_link == element_number)
        else:
            index = None
        assert index, 'Element number is not a channel link'

        x_location = d.x_coordinates[int(index[1])]
        y_location = d.y_coordinates[int(index[0])]

        return x_location,y_location

    def get_element_location(self, dem_file, element_number):
        d = Dem(dem_file)

        index = np.where(self.number.square == element_number)

        x_location = d.x_coordinates[int(index[1])]
        y_location = d.y_coordinates[int(index[0])]

        return x_location, y_location

    def to_json(self):

        d = {}

        for key in self.__dict__.keys():

            attr = self.__dict__[key]

            if type(attr) == Constant:
                d[key] = {}
                for constant_key in attr.__dict__.keys():
                    constant_attr = attr.__dict__[constant_key]
                    d[key][constant_key] = constant_attr.tolist()

            elif type(attr) == Variable:
                d[key] = {}
                d[key]['values'] = attr.values[:].tolist()
                d[key]['times'] = attr.times[:].tolist()

        return d

    def to_geom(self, dem, srs='EPSG:27700'):

        dem = Dem(dem)

        features = []

        geoms = Geometries(self, dem, srs)

        for n in self.element_numbers:


            properties = {}

            if self.overland_flow:
                properties['overland_flow'] = {
                    'values':(np.absolute(self.overland_flow.values[n-1,:,:]).max(axis=0).astype(np.float64).tolist()
                              if n-1<self.overland_flow.values.shape[0] else [])
                         }
            if self.ph_depth:
                properties['ph_depth'] = {
                    'values':self.ph_depth.values[:][self.number.square==n].flatten().tolist()
                }

            if self.surface_depth:
                properties['surface_depth'] = {
                    'values':(self.surface_depth.values[n-1,:].astype(np.float64).tolist()
                              if n-1<self.overland_flow.values.shape[0] else [])
                         }

            if self.canopy_storage:
                properties['canopy_storage'] = {
                    'values':self.canopy_storage.values[:][self.number.square==n].flatten().tolist()
                }
            if self.soil_moisture:
                properties['theta'] = {
                    'values': self.soil_moisture.values[:, :, 0, 1:][self.number.square == n].flatten().tolist()
                }

            properties['dem'] = {
                'value': float(self.sv4_elevation[self.sv4_numbering == n][0])
            }

            properties['number'] = int(n)


            features.append({
                'type': 'Feature',
                'geometry': geoms.__next__(),
                'properties': properties
            })

        # unique, inverse = np.unique(numbers, return_inverse=True)


        variables = []

        if self.ph_depth:
            variables.append(
                {
                'name': 'ph_depth',
                'longName':'Phreatic Depth (m)',
                'max': self.ph_depth.values[:][self.ph_depth.values[:] != -1].astype(np.float64).max(),
                'min': self.ph_depth.values[:][self.ph_depth.values[:] != -1].astype(np.float64).min(),
                'times': self.ph_depth.times[:].tolist()
                })

        if self.overland_flow:
            variables.append(
                {
                'name': 'overland_flow',
                'longName':'Overland Flow (cumecs)',
                'max': np.absolute(self.overland_flow.values[:].astype(np.float64)).max(),
                'min': np.absolute(self.overland_flow.values[:].astype(np.float64)).min(),
                'times':self.overland_flow.times[:].tolist()
                })
        if self.canopy_storage:
            variables.append(
                {
                'name': 'canopy_storage',
                'longName':'Canopy Storage (mm)',
                'max': self.canopy_storage.values[:][self.canopy_storage.values[:] != -1].astype(np.float64).max(),
                'min': self.canopy_storage.values[:][self.canopy_storage.values[:] != -1].astype(np.float64).min(),
                'times': self.canopy_storage.times[:].tolist()
                })
        if self.surface_depth:
            variables.append(
                {
                    'name': 'surface_depth',
                    'longName': 'Surface Depth (m)',
                    'max': self.surface_depth.values[:].astype(np.float64).max(),
                    'min': self.surface_depth.values[:].astype(np.float64).min(),
                    'times': self.surface_depth.times[:].tolist()
                }
            )
        if self.soil_moisture:
            variables.append(
                {
                    'name': 'theta',
                    'longName': 'Soil Moisture (m3/m3)',
                    'max': (self.soil_moisture.values[:, :, 0, 1:][self.soil_moisture.values[:, :, 0, 1:] != -1].astype(np.float64).max()
                        if len(self.soil_moisture.values[:, :, 0, 1:][self.soil_moisture.values[:, :, 0, 1:] != -1]) > 0 else None),
                    'min': (self.soil_moisture.values[:, :, 0, 1:][self.soil_moisture.values[:, :, 0, 1:] != -1].astype(np.float64).min()
                        if len(self.soil_moisture.values[:, :, 0, 1:][self.soil_moisture.values[:, :, 0, 1:] != -1]) > 0 else None),
                    'times': self.soil_moisture.times[1:].tolist()
                }
            )


        return {
            'geom':
                    {
                        'type':'FeatureCollection',
                        'features':features
                    },
            'variables':variables
        }


class Geometries:
    def __init__(self, hdf, dem, srs=None):
        from osgeo import osr

        self.hdf = hdf
        self.dem = dem
        if srs is None:
            srs = 'EPSG:27700'
            print('SRID not specified, setting to {}'.format(srs))

        assert ':' in srs, 'SRID must be formatted like EPSG:27700'

        authority, code = srs.split(':')
        source = osr.SpatialReference()

        if authority == 'EPSG':
            source.ImportFromEPSG(int(code))
        else:
            raise Exception('Only EPSG coordinate systems are currently supported')

        target = osr.SpatialReference()
        target.ImportFromEPSG(4326)

        self.transform = osr.CoordinateTransformation(source, target)

        n = hdf.sv4_numbering

        self.indices = np.indices(n.shape)

        cell_size_factor = hdf.sv4_elevation.shape[0] / hdf.surface_elevation.square.shape[0]

        self.cell_size = dem.cell_size / cell_size_factor

        self.current = 0

        _, self.index = np.unique(n.flatten(), return_index=True)
        _, self.reverse_index = np.unique(n.flatten()[::-1], return_index=True)

        self.y = (self.indices[0] * self.cell_size + dem.y_lower_left - dem.cell_size)[::-1]
        self.x = self.indices[1] * self.cell_size + dem.x_lower_left - dem.cell_size

        self.ymax = self.y.flatten()[::-1][self.reverse_index][1:]
        self.ymin = self.y.flatten()[self.index][1:]

        self.xmax = self.x.flatten()[::-1][self.reverse_index][1:]
        self.xmin = self.x.flatten()[self.index][1:]

    def __len__(self):
        return len(self.ymax)

    def __iter__(self):
        return self

    def __next__(self):
        from osgeo import ogr
        if self.current > len(self) - 1:
            raise StopIteration
        else:
            y1 = self.ymin[self.current]
            x1 = self.xmin[self.current]
            y2 = self.ymax[self.current]
            x2 = self.xmax[self.current]

            point = ogr.CreateGeometryFromWkt("POINT ({} {})".format(x1, y1))
            point.Transform(self.transform)

            x1 = point.GetX()
            y1 = point.GetY()

            point = ogr.CreateGeometryFromWkt("POINT ({} {})".format(x2, y2))
            point.Transform(self.transform)

            x2 = point.GetX()
            y2 = point.GetY()

            self.current += 1

            return {
                'type': 'Polygon',
                'coordinates': [[[x1, y1], [x1, y2], [x2, y2], [x2, y1], [x1, y1]]]
            }

variable_names = {
    'net_rain': 'Net Rain',
    'trnsp': 'Transpiration',
    'pot_evap': 'Potential Evapotranspiration',
    'srf_evap': 'Surface Evaporation',
    'int_evap': 'Evaporation from Interception',
    'drainage': 'Drainage from Interception',
    'can_stor': 'Canopy Storage',
    'v_flow': 'Vertical Flows',
    'snow_dep': 'Snow Depth',
    'ph_depth': 'Phreatic Depth',
    'ovr_flow': 'Overland Flow',
    'srf_dep': 'Surface Depth',
    'psi': 'Surface Water Potential',
    'theta': 'Soil Moisture',
    's_t_dp': 'Total Sediment Depth',
    's_v_er': 'Surface Erosion Rate',
    's_dis': 'Sediment Discharge Rate',
    'bal_err': 'Mass Balance Error'
}




import os
import shutil
import subprocess
from datetime import *
from .run import Run
import numpy as np
import gdal
from . import gear

from . import mask
from .library import make_lib_file

rainfall_source = 'GEAR'


def get_temp(pe_id_list, start_date, end_date, temp_time_series_file, max_or_min):
    # calculate start and end line numbers
    start_line_delta = datetime.strptime(start_date, "%Y-%m-%d") - datetime.strptime("01/01/1960", "%d/%m/%Y")
    lines_to_skip = start_line_delta.days
    lines_to_read_delta = datetime.strptime(end_date, "%Y-%m-%d") - datetime.strptime(start_date, "%Y-%m-%d")
    lines_to_read = lines_to_read_delta.days  # this doesn't count the end date... i think this is what SHETRAN wants
    # get time series for each id
    all_grid_squares_time_series_list = []
    for i in range(len(pe_id_list)):
        one_grid_square_time_series_list = [str(pe_id_list[i])]
        time_series_id = int(pe_id_list[i])
        # get numbers of the equivalent file
        if time_series_id == -9999:
            pass
        else:
            if time_series_id % 180 == 0:
                x_coord = 180
                y_coord = int(round(time_series_id / 180, 0))
            else:
                x_coord = time_series_id % 180
                y_coord = int(round(time_series_id / 180, 0)) + 1

            assert isinstance(max_or_min, str)
            temp_file = open("../UKCP09" + max_or_min + "TempTimeSeries/[" + str(x_coord) + "][" + str(
                y_coord) + "]" + max_or_min + "TempTimeSeries.txt", "r")

            for a in range(lines_to_skip):
                temp_file.readline()
            for b in range(lines_to_read):
                temp_val = temp_file.readline().rstrip()
                one_grid_square_time_series_list.append(temp_val)

            temp_file.close()
            all_grid_squares_time_series_list.append(one_grid_square_time_series_list)

    print("writing temp time series data to file")
    for c in range(len(all_grid_squares_time_series_list[0])):
        line_list = []
        for d in range(len(all_grid_squares_time_series_list)):
            line_list.append(str(all_grid_squares_time_series_list[d][c]))
        line = ",".join(line_list)
        if c == 0:
            temp_time_series_file.write(line + "," + start_date + " to " + end_date + "\n")
        else:
            temp_time_series_file.write(line + "\n")


def setup(run: Run):

    directory = "outputs/{}".format(run.gauge_id)
    print("Creating directory...")
    if not os.path.exists(directory):
        os.makedirs(directory)

    if run.boundary is not None:
        mask.create(run.outline, run.resolution, out)
    else:
        shutil.copy2(run.inputs.mask, run.outputs.mask)

    mask_file = open(run.outputs.mask, "r")

    n_cols_line = mask_file.readline()
    n_cols_list = n_cols_line.rstrip().split()
    n_cols = float(n_cols_list[1])

    n_rows_line = mask_file.readline()
    n_rows_list = n_rows_line.rstrip().split()
    n_rows = float(n_rows_list[1])

    bx_line = mask_file.readline()
    bx_list = bx_line.rstrip().split()
    x_lower_left_corner = float(bx_list[1])

    by_line = mask_file.readline()
    by_list = by_line.rstrip().split()
    y_lower_left_corner = float(by_list[1])

    mask_file.close()

    def extract(in_file, out_file):
        ds = gdal.Open(in_file)

        start_line_number = ds.RasterYSize - y_lower_left_corner / run.resolution_in_metres - n_rows
        start_index_number = x_lower_left_corner / run.resolution_in_metres

        gdal.Translate(out_file, ds, srcWin=[
                start_index_number,
                start_line_number,
                n_cols,
                n_rows],
                       format='AAIGrid')
        del ds
        ds = gdal.Open(out_file)
        unique_values = np.unique(ds.ReadAsArray())
        del ds
        return unique_values

    extract(run.inputs.dem, run.outputs.dem)
    extract(run.inputs.min_dem, run.outputs.min_dem)
    extract(run.inputs.soil, run.outputs.soil)
    extract(run.inputs.veg, run.outputs.veg)
    extract(run.inputs.lakes, run.outputs.lakes)

    pe_output = extract(run.inputs.pe, run.outputs.pe)

    if rainfall_source == 'GEAR':
        gear.extract(run)
    else:
        rain_output = extract(run.inputs.rain, run.outputs.rain)

        unique_rain_ids_list = list(np.sort(rain_output[rain_output != -9999]).astype(str))

        # calculate start and end line numbers
        start_line_delta = datetime.strptime(run.start_date, "%Y-%m-%d") - datetime.strptime("01/01/1958", "%d/%m/%Y")
        lines_to_skip = start_line_delta.days
        lines_to_read_delta = datetime.strptime(run.end_date, "%Y-%m-%d") - datetime.strptime(run.start_date, "%Y-%m-%d")
        lines_to_read = lines_to_read_delta.days  # this doesn't count the end date... i think this is what SHETRAN wants
        # get time series for each id
        all_grid_squares_time_series_list = []
        print(unique_rain_ids_list)
        for i in range(len(unique_rain_ids_list)):
            one_grid_square_time_series_list = [str(unique_rain_ids_list[i])]
            time_series_id = int(unique_rain_ids_list[i])
            # get numbers of the equivalent file
            if time_series_id % 180 == 0:
                x_coord = 180
                y_coord = time_series_id / 180
            else:
                x_coord = time_series_id % 180
                y_coord = int((time_series_id / 180) + 1)

            rain_file = open(
                "../Inputs/UKCP09RainfallTimeSeries/[" + str(x_coord) + "][" + str(y_coord) + "]RainfallTimeSeries.txt",
                "r")

            for a in range(lines_to_skip):
                rain_file.readline()
            for b in range(lines_to_read):
                rain_val = rain_file.readline().rstrip()
                one_grid_square_time_series_list.append(rain_val)

            rain_file.close()
            all_grid_squares_time_series_list.append(one_grid_square_time_series_list)

        print("writing Rain time series data to file")
        with open(run.outputs.rain_timeseries, 'w') as f:
            for c in range(len(all_grid_squares_time_series_list[0])):
                line_list = []
                for d in range(len(all_grid_squares_time_series_list)):
                    line_list.append(str(all_grid_squares_time_series_list[d][c]))
                line = ",".join(line_list)
                if c == 0:
                    f.write(line + "," + run.start_date + " to " + run.end_date + "\n")
                else:
                    f.write(line + "\n")

    unique_p_e_ids_list = np.sort(pe_output[pe_output != -9999]).astype(str)

    new_p_e_time_series = open(run.outputs.pe_timeseries, 'w')

    # calculate start and end line numbers
    start_line_delta = datetime.strptime(run.start_date, "%Y-%m-%d") - datetime.strptime("01/01/1960", "%d/%m/%Y")
    lines_to_skip = start_line_delta.days
    lines_to_read_delta = datetime.strptime(run.end_date, "%Y-%m-%d") - datetime.strptime(run.start_date, "%Y-%m-%d")
    lines_to_read = lines_to_read_delta.days  # this doesn't count the end date... i think this is what SHETRAN wants
    # get time series for each id
    all_grid_squares_time_series_list = []
    for i in range(len(unique_p_e_ids_list)):
        one_grid_square_time_series_list = [str(unique_p_e_ids_list[i])]
        time_series_id = int(unique_p_e_ids_list[i])
        # get numbers of the equivalent file
        if time_series_id % 180 == 0:
            x_coord = 180
            y_coord = time_series_id / 180
        else:
            x_coord = time_series_id % 180
            y_coord = int((time_series_id / 180) + 1)

        pe_file = open(
            "../UKCP09PEBetterTimeSeries/[" + str(x_coord) + "][" + str(y_coord) + "]PEBetterTimeSeries.txt",
            "r")

        for a in range(lines_to_skip):
            pe_file.readline()
        for b in range(lines_to_read):
            pe_val = pe_file.readline().rstrip()
            one_grid_square_time_series_list.append(pe_val)

        pe_file.close()
        all_grid_squares_time_series_list.append(one_grid_square_time_series_list)

        # write to file
    print("writing PE time series data to file")
    for c in range(len(all_grid_squares_time_series_list[0])):
        line_list = []
        for d in range(len(all_grid_squares_time_series_list)):
            line_list.append(str(all_grid_squares_time_series_list[d][c]))
        line = ",".join(line_list)
        if c == 0:
            new_p_e_time_series.write(line + "," + run.start_date + " to " + run.end_date + "\n")
        else:
            new_p_e_time_series.write(line + "\n")

    new_p_e_time_series.close()

    get_temp(unique_p_e_ids_list, run.start_date, run.end_date,
             open(run.outputs.max_temp_timeseries, 'w'),
             "Max")
    get_temp(unique_p_e_ids_list, run.start_date, run.end_date,
             open(run.outputs.min_temp_timeseries, "w"),
             "Min")
    # write time series data to file

    # write library file

    make_lib_file(run)

    # copy SHETRAN files over

    print("Copying SHETRAN files over")
    if not os.path.exists(directory + "/program"):
        shutil.copytree("../program", directory + "/program")
    if not os.path.exists(directory+"/prepare-liz2.2.4a.exe"):
        shutil.copy("../prepare-liz2.2.4a.exe", directory + "/")
    if not os.path.exists(directory+"/rdf.csv"):
        shutil.copy("../rdf.csv", directory + "/")

    # time.sleep(10)

    # run SHETRAN prepare
    # catch = raw_input(enter catch again)

    # directory = os.getcwd()
    # dirList = directory.split("\\")
    # directory2 = "\\".join(dirList)
    # print(catch)
    # print(os.getcwd())
    #
    # new_directory = directory2 + "\\outputs\\" + catch
    # if not os.path.exists(new_directory):
    #     os.mkdir(new_directory)
    # os.chdir(new_directory)
    # print(os.getcwd())
    # libFile = catch + "LibraryFile.xml"
    # subprocess.call(["prepare-liz2.2.4a.exe", libFile])


def run_shetran(catch):
    os.chdir("outputs\\" + catch + "\\program")
    run_file = "outputs\\" + catch + "\\rundata_" + catch + ".txt"
    subprocess.call(["sv4.4.2asnow.exe", "-f", run_file])


def get_observed_flows(run):
    # obs_file = open("E:/SHETRANforUK/Inputs/NRFAdata/ "+ str(gauge)  + "FlowData.csv", "r")
    start2 = run.start_date.replace("-", ".")
    end2 = run.end_date.replace("-", ".")

    new_file = open(run.outputs.observed_flow, "w")

    obs_path = "../Inputs/NRFAdata/" + str(run.gauge_id) + "FlowData.csv"
    if os.path.exists(obs_path):
        obs_file = open(obs_path, "r")
    else:
        print('There is no flow data available for gauge {gauge_id}'.format(gauge_id=run.gauge_id))
        return

    for x in range(14):
        obs_file.readline()

    obs_start_list = obs_file.readline().rstrip().split(",")
    obs_start_string = obs_start_list[2]
    obs_start = datetime.strptime(obs_start_string, "%Y-%m-%d")
    print(obs_start)

    obs_end_list = obs_file.readline().rstrip().split(",")
    obs_end_string = obs_end_list[2]
    obs_end = datetime.strptime(obs_end_string, "%Y-%m-%d")
    print(obs_end)

    start_date = datetime.strptime(run.start_date, "%Y-%m-%d")
    end_date = datetime.strptime(run.end_date, "%Y-%m-%d")
    print(start_date)
    print(end_date)

    new_file.write("date,flow\n")

    if (obs_start - start_date).days > 0:
        print(
            "I'm sorry, I don't have observed flows for the period you're modelling. I only have data for " +
            obs_start_string + " to " + obs_end_string)
    elif (end_date - obs_end).days > 0:
        print(
            "I'm sorry, I don't have observed flows for the period you're modelling. I only have data for " +
            obs_start_string + " to " + obs_end_string)
    else:
        days_to_skip = (start_date - obs_start).days
        print("days to skip = " + str(days_to_skip))

        for x in range(days_to_skip):
            obs_file.readline()

        days_to_read = (end_date - start_date).days
        print("days to read = " + str(days_to_read))
        for x in range(days_to_read):
            line_list = obs_file.readline().rstrip().split(",")
            new_list = line_list[:2]
            new_line = ",".join(new_list)
            new_file.write(new_line + "\n")

    new_file.close()


def setup_catchment(run):
    setup(run)
    directory = "outputs/" + str(run.gauge_id)
    if run.upload is None:
        get_observed_flows(run)
    import zipfile
    print('writing zip file')
    zf = zipfile.ZipFile('output.zip', 'w', zipfile.ZIP_DEFLATED)
    print(os.path.exists('outputs/' + str(run.gauge_id)))
    for root, dirs, files in os.walk('outputs/' + str(run.gauge_id)):
        for file in files:
            print(file)
            zf.write(os.path.join(root, file))

    zf.close()



import subprocess
import os
from datetime import datetime
import netCDF4 as nc
import gdal
import numpy as np


def download_ceh_gear(username, password, output_directory=None, start=1890, end=2016):
    base_url = "https://catalogue.ceh.ac.uk/datastore/eidchub/33604ea0-c238-4488-813d-0ad9ab7c51ca/GB/daily/"

    for year in range(start, end):

        filename = "CEH_GEAR_daily_GB_{}.nc".format(year)

        url = base_url + filename

        if output_directory is not None:
            output_file = os.path.join(output_directory, filename)
        else:
            output_file = filename

        subprocess.call(["curl", "--user", "{}:{}".format(username, password),  url, "-o", output_file])


def extract(data_path: str,
            variable: str,
            mask_path: str,
            start_date: datetime,
            end_date: datetime,
            output_grid_path: str,
            output_ts_path: str) -> None:

    mask = gdal.Open(mask_path)

    x_min, x_res, x_skew, y_max, y_skew, y_res = mask.GetGeoTransform()

    mask_width = mask.RasterXSize
    mask_height = mask.RasterYSize

    x_max = x_min + (mask_width * x_res)
    y_min = y_max + (mask_height * y_res)
    no_data = mask.GetRasterBand(1).GetNoDataValue()

    mask_x = np.arange(x_min, x_max + x_res, x_res)
    mask_y = np.arange(y_min, y_max + x_res, x_res)

    mask_y_idx, mask_x_idx = np.where(mask.ReadAsArray()!=no_data)

    mask_y_selected = mask_y[mask_y_idx]
    mask_x_selected = mask_x[mask_x_idx]


    ds = nc.Dataset(data_path)

    data_x = ds.variables['x'][:]
    data_y = ds.variables['y'][:]

    data_y_selected = [data_y[data_y <= y_val].max() for y_val in mask_y_selected]
    data_x_selected = [data_x[data_x <= x_val].max() for x_val in mask_x_selected]

    data_y_mask = np.isin(data_y, data_y_selected)
    data_x_mask = np.isin(data_x, data_x_selected)

    time = ds.variables['time']

    dates = nc.num2date(time[:], time.units, time.calendar)
    dates_mask = (dates>=start_date)&(dates<=end_date)

    values = ds.variables[variable][dates_mask, data_y_mask, data_x_mask]

    data_unique_x = np.unique(data_y_selected).astype(int)
    data_unique_y = np.unique(data_x_selected).astype(int)

    output = np.full((mask_height, mask_width), no_data).astype(int)

    series = []
    cells = {}
    number = 1

    for i in range(len(mask_x_selected)):

        data_y = data_y_selected[i]
        data_x = data_x_selected[i]
        x = mask_x_selected[i]
        y = mask_y_selected[i]

        if (data_x, data_y) not in cells:
            series.append( values[:, np.where(data_unique_x == data_y)[0][0], np.where(data_unique_y == data_x)[0][0]])
            cells[(data_x, data_y)] = number
            number += 1

        output[np.where(mask_y==y), np.where(mask_x==x)] = cells[(data_x, data_y)]

    driver = gdal.GetDriverByName("GTiff")
    tif_path = output_grid_path + '.tif'
    grid = driver.CreateCopy(tif_path, mask, 0)
    grid.GetRasterBand(1).WriteArray(output)

    driver = gdal.GetDriverByName("AAIGrid")
    driver.CreateCopy(output_grid_path, grid, 0)

    with open(output_ts_path, 'w') as f:
        f.write(','.join(np.arange(1, len(series)+1).astype(str))+'\n')
        for i in range(len(series[0])):
            f.write(','.join([str(cell_series[i]) for cell_series in series])+'\n')
